"""
Theming service.

Themes are stored in the css folder, current theme in UserSetting.
"""

import os
from glob import glob
from flask import current_app
from lute.models.repositories import UserSettingRepository

default_entry = ("-", "(default)")


class Service:
    "Service."

    def __init__(self, session):
        self.session = session

    def _css_path(self):
        """
        Path to css in this folder.
        """
        thisdir = os.path.dirname(__file__)
        theme_dir = os.path.join(thisdir, "css")
        return os.path.abspath(theme_dir)

    def list_themes(self):
        """
        List of theme file names and user-readable name.
        """

        def _make_display_name(s):
            ret = os.path.basename(s)
            ret = ret.replace(".css", "").replace("_", " ")
            return ret

        g = glob(os.path.join(self._css_path(), "*.css"))
        themes = [(os.path.basename(f), _make_display_name(f)) for f in g]
        theme_basenames = [t[0] for t in themes]

        g = glob(os.path.join(current_app.env_config.userthemespath, "*.css"))
        additional_user_themes = [
            (os.path.basename(f), _make_display_name(f))
            for f in g
            if os.path.basename(f) not in theme_basenames
        ]

        themes += additional_user_themes
        sorted_themes = sorted(themes, key=lambda x: x[1])
        return [default_entry] + sorted_themes

    def get_current_css(self):
        """
        Return the current css pointed at by the current_theme user setting.
        """
        repo = UserSettingRepository(self.session)
        current_theme = repo.get_value("current_theme")
        if current_theme == default_entry[0]:
            return ""

        def _get_theme_css_in_dir(d):
            "Get css, or '' if no file."
            fname = os.path.join(d, current_theme)
            if not os.path.exists(fname):
                return ""
            with open(fname, "r", encoding="utf-8") as f:
                return f.read()

        ret = _get_theme_css_in_dir(self._css_path())
        add = _get_theme_css_in_dir(current_app.env_config.userthemespath)
        if add != "":
            ret += f"\n\n/* Additional user css */\n\n{add}"
        return ret

    def next_theme(self):
        """
        Move to the next theme in the list of themes.
        """
        repo = UserSettingRepository(self.session)
        current_theme = repo.get_value("current_theme")
        themes = [t[0] for t in self.list_themes()]
        themes.append(default_entry[0])
        for i in range(0, len(themes)):  # pylint: disable=consider-using-enumerate
            if themes[i] == current_theme:
                new_index = i + 1
                break
        repo.set_value("current_theme", themes[new_index])
        self.session.commit()



"Data cleanup tests."

from datetime import datetime
from sqlalchemy import text as sqltext
from lute.db import db
from lute.db.data_cleanup import clean_data
from tests.utils import make_text
from tests.dbasserts import assert_sql_result


# Cleaning up missing sentence.SeTextLC records.


def test_cleanup_loads_missing_sentence_textlc(app_context, spanish):
    """
    Load the sentence.SeTextLC.

    If the sqlite LOWER(SeText) would be the same as the parser-generated lowercase text,
    store the special char '*' only, don't waste file space storing the parser-generated lc text.
    """

    t = make_text("test", "gato. Ábrelo. tengo. QUIERO. Ábrela. ábrela.", spanish)
    t.read_date = datetime.now()
    db.session.add(t)
    db.session.commit()

    # Force re-calc.
    sqlhack = """
    update sentences set SeTextLC = null
    where SeText not like '%gato%' and SeText not like '%brelo%'
    """
    db.session.execute(sqltext(sqlhack))
    db.session.commit()
    sql = "select SeText, SeTextLC from sentences order by SeID"
    preclean = [
        "/gato/./; *",
        "/Ábrelo/./; /ábrelo/./",
        "/tengo/./; None",
        "/QUIERO/./; None",
        "/Ábrela/./; None",
        "/ábrela/./; None",
    ]
    assert_sql_result(sql, preclean, "pre-clean")

    def _output(s):
        print(s, flush=True)

    clean_data(db.session, _output)

    postclean = [
        "/gato/./; *",
        "/Ábrelo/./; /ábrelo/./",
        "/tengo/./; *",
        "/QUIERO/./; *",
        "/Ábrela/./; /ábrela/./",
        "/ábrela/./; *",
    ]
    assert_sql_result(sql, postclean, "post-clean")



"""
/read endpoints.
"""

from flask import Blueprint, flash, request, render_template, redirect, jsonify
from lute.read.service import Service
from lute.read.forms import TextForm
from lute.term.model import Repository
from lute.term.routes import handle_term_form
from lute.settings.current import current_settings
from lute.models.book import Text
from lute.models.repositories import BookRepository, LanguageRepository
from lute.db import db


bp = Blueprint("read", __name__, url_prefix="/read")


def _render_book_page(book, pagenum):
    """
    Render a particular book page.
    """
    lang = book.language
    show_highlights = current_settings["show_highlights"]
    lang_repo = LanguageRepository(db.session)
    term_dicts = lang_repo.all_dictionaries()[lang.id]["term"]

    return render_template(
        "read/index.html",
        hide_top_menu=True,
        is_rtl=lang.right_to_left,
        html_title=book.title,
        book=book,
        sentence_dict_uris=lang.sentence_dict_uris,
        page_num=pagenum,
        page_count=book.page_count,
        show_highlights=show_highlights,
        lang_id=lang.id,
        term_dicts=term_dicts,
    )


def _find_book(bookid):
    "Find book from db."
    br = BookRepository(db.session)
    return br.find(bookid)


@bp.route("/<int:bookid>", methods=["GET"])
def read(bookid):
    """
    Read a book, opening to its current page.

    This is called from the book listing, on Lute index.
    """
    book = _find_book(bookid)
    if book is None:
        flash(f"No book matching id {bookid}")
        return redirect("/", 302)

    page_num = 1
    text = book.texts[0]
    if book.current_tx_id:
        text = db.session.get(Text, book.current_tx_id)
        page_num = text.order

    return _render_book_page(book, page_num)


@bp.route("/<int:bookid>/page/<int:pagenum>", methods=["GET"])
def read_page(bookid, pagenum):
    """
    Read a particular page of a book.

    Called from term Sentences link.
    """
    book = _find_book(bookid)
    if book is None:
        flash(f"No book matching id {bookid}")
        return redirect("/", 302)

    pagenum = book.page_in_range(pagenum)
    return _render_book_page(book, pagenum)


@bp.route("/page_done", methods=["post"])
def page_done():
    "Handle POST when page is done."
    data = request.json
    bookid = int(data.get("bookid"))
    pagenum = int(data.get("pagenum"))
    restknown = data.get("restknown")

    service = Service(db.session)
    service.mark_page_read(bookid, pagenum, restknown)
    return jsonify("ok")


@bp.route("/delete_page/<int:bookid>/<int:pagenum>", methods=["GET"])
def delete_page(bookid, pagenum):
    """
    Delete page.
    """
    book = _find_book(bookid)
    if book is None:
        flash(f"No book matching id {bookid}")
        return redirect("/", 302)

    if len(book.texts) == 1:
        flash("Cannot delete only page in book.")
    else:
        book.remove_page(pagenum)
        db.session.add(book)
        db.session.commit()

    url = f"/read/{bookid}/page/{pagenum}"
    return redirect(url, 302)


@bp.route("/new_page/<int:bookid>/<position>/<int:pagenum>", methods=["GET", "POST"])
def new_page(bookid, position, pagenum):
    "Create a new page."
    form = TextForm()
    book = _find_book(bookid)

    if form.validate_on_submit():
        t = None
        if position == "before":
            t = book.add_page_before(pagenum)
        else:
            t = book.add_page_after(pagenum)
        t.book = book
        t.text = form.text.data
        db.session.add(book)
        db.session.commit()

        book.current_tx_id = t.id
        db.session.add(book)
        db.session.commit()

        return redirect(f"/read/{book.id}", 302)

    text_dir = "rtl" if book.language.right_to_left else "ltr"
    return render_template(
        "read/page_edit_form.html", hide_top_menu=True, form=form, text_dir=text_dir
    )


@bp.route("/save_player_data", methods=["post"])
def save_player_data():
    "Save current player position, bookmarks.  Called on a loop by the player."
    data = request.json
    bookid = int(data.get("bookid"))
    book = _find_book(bookid)
    book.audio_current_pos = float(data.get("position"))
    book.audio_bookmarks = data.get("bookmarks")
    db.session.add(book)
    db.session.commit()
    return jsonify("ok")


@bp.route("/renderpage/<int:bookid>/<int:pagenum>", methods=["GET"])
def render_page(bookid, pagenum):
    "Method called by ajax, render the given page."
    book = _find_book(bookid)
    if book is None:
        flash(f"No book matching id {bookid}")
        return redirect("/", 302)
    service = Service(db.session)
    paragraphs = service.start_reading(book, pagenum)
    return render_template("read/page_content.html", paragraphs=paragraphs)


@bp.route("/empty", methods=["GET"])
def empty():
    "Show an empty/blank page."
    return ""


@bp.route("/termform/<int:langid>/<text>", methods=["GET", "POST"])
def term_form(langid, text):
    """
    Create a multiword term for the given text, replacing the LUTESLASH hack.
    """
    usetext = text.replace("LUTESLASH", "/")
    repo = Repository(db.session)
    term = repo.find_or_new(langid, usetext)
    if term.status == 0:
        term.status = 1
    return handle_term_form(
        term,
        repo,
        db.session,
        "/read/term_edit_form.html",
        render_template("/read/updated.html", term_text=term.text),
        embedded_in_reading_frame=True,
    )


@bp.route("/edit_term/<int:term_id>", methods=["GET", "POST"])
def edit_term_form(term_id):
    """
    Edit a term.
    """
    repo = Repository(db.session)
    term = repo.load(term_id)
    # print(f"editing term {term_id}", flush=True)
    if term.status == 0:
        term.status = 1
    return handle_term_form(
        term,
        repo,
        db.session,
        "/read/term_edit_form.html",
        render_template("/read/updated.html", term_text=term.text),
        embedded_in_reading_frame=True,
    )


@bp.route("/term_bulk_edit_form", methods=["GET"])
def term_bulk_edit_form():
    """
    show_bulk_form
    """
    repo = Repository(db.session)
    return render_template(
        "read/term_bulk_edit_form.html",
        tags=repo.get_term_tags(),
    )


@bp.route("/termpopup/<int:termid>", methods=["GET"])
def term_popup(termid):
    """
    Get popup html for DBTerm, or None if nothing should be shown.
    """
    service = Service(db.session)
    d = service.get_popup_data(termid)
    if d is None:
        return ""
    return render_template(
        "read/termpopup.html",
        data=d,
    )


@bp.route("/flashcopied", methods=["GET"])
def flashcopied():
    return render_template("read/flashcopied.html")


@bp.route("/editpage/<int:bookid>/<int:pagenum>", methods=["GET", "POST"])
def edit_page(bookid, pagenum):
    "Edit the text on a page."
    book = _find_book(bookid)
    text = book.text_at_page(pagenum)
    if text is None:
        return redirect("/", 302)
    form = TextForm(obj=text)

    if form.validate_on_submit():
        form.populate_obj(text)
        db.session.add(text)
        db.session.commit()
        return redirect(f"/read/{book.id}", 302)

    text_dir = "rtl" if book.language.right_to_left else "ltr"
    return render_template(
        "read/page_edit_form.html", hide_top_menu=True, form=form, text_dir=text_dir
    )



"""
Settings routes.
"""

from flask import (
    Blueprint,
    current_app,
    request,
    render_template,
    redirect,
    flash,
    jsonify,
)
from wtforms import BooleanField
from lute.models.language import Language
from lute.models.setting import UserSetting
from lute.models.repositories import UserSettingRepository
from lute.themes.service import Service as ThemeService
from lute.settings.forms import UserSettingsForm, UserShortcutsForm
from lute.settings.current import refresh_global_settings
from lute.db import db
from lute.parse.mecab_parser import JapaneseParser


bp = Blueprint("settings", __name__, url_prefix="/settings")


@bp.route("/index", methods=["GET", "POST"])
def edit_settings():
    "Edit settings."
    form = UserSettingsForm()

    with current_app.app_context():
        svc = ThemeService(db.session)
        form.current_theme.choices = svc.list_themes()

    ac = current_app.env_config
    if ac.is_docker:
        # User shouldn't change some things with docker.
        kw = {"readonly": True, "style": "background-color: LightGray"}
        # Backup dir gets mounted from host.
        form.backup_dir.render_kw = kw

    repo = UserSettingRepository(db.session)
    if form.validate_on_submit():
        # Update the settings in the database
        for field in form:
            if field.id not in ("csrf_token", "submit"):
                repo.set_value(field.id, field.data)
        db.session.commit()
        refresh_global_settings(db.session)

        flash("Settings updated", "success")
        return redirect("/")

    # Load current settings from the database
    for field in form:
        if field.id != "csrf_token":
            field.data = repo.get_value(field.id)
        if isinstance(field, BooleanField):
            # Hack: set boolean settings to ints, otherwise they're always checked.
            field.data = int(field.data or 0)

    return render_template("settings/form.html", form=form)


@bp.route("/test_mecab", methods=["GET"])
def test_parse():
    """
    Do a test parse for the JapaneseParser using the
    given path string.

    Returns { 'success': tokens }, or { 'error' msg }

    """
    mecab_path = request.args.get("mecab_path", None)
    repo = UserSettingRepository(db.session)
    old_setting = repo.get_value("mecab_path")
    result = {"failure": "tbd"}
    try:
        repo.set_value("mecab_path", mecab_path)
        # Parsing requires a language, even if it's a dummy.
        lang = Language()
        p = JapaneseParser()
        src = "私は元気です"
        toks = p.get_parsed_tokens(src, lang)
        toks = [tok.token for tok in toks if tok.token != "¶"]
        message = f"{src} parsed to [{ ', '.join(toks) }]"
        result = {"result": "success", "message": message}
    except Exception as e:  # pylint: disable=broad-exception-caught
        message = f"{type(e).__name__}: { str(e) }"
        result = {"result": "failure", "message": message}
    finally:
        repo.set_value("mecab_path", old_setting)

    return jsonify(result)


@bp.route("/set/<key>/<value>", methods=["POST"])
def set_key_value(key, value):
    "Set a UserSetting key to value."
    repo = UserSettingRepository(db.session)
    old_value = repo.get_value(key)
    try:
        repo.set_value(key, value)
        result = {"result": "success", "message": "OK"}
    except Exception as e:  # pylint: disable=broad-exception-caught
        message = f"{type(e).__name__}: { str(e) }"
        repo.set_value(key, old_value)
        result = {"result": "failure", "message": message}
    db.session.commit()
    refresh_global_settings(db.session)
    return jsonify(result)


def _get_categorized_hotkeys():
    """
    Return hotkey UserSetting keys and values,
    grouped by category.
    """

    categorized_settings = {
        "Navigation": [
            "hotkey_StartHover",
            "hotkey_PrevWord",
            "hotkey_NextWord",
            "hotkey_PrevUnknownWord",
            "hotkey_NextUnknownWord",
            "hotkey_PrevSentence",
            "hotkey_NextSentence",
        ],
        "Update status": [
            "hotkey_Status1",
            "hotkey_Status2",
            "hotkey_Status3",
            "hotkey_Status4",
            "hotkey_Status5",
            "hotkey_StatusIgnore",
            "hotkey_StatusWellKnown",
            "hotkey_StatusUp",
            "hotkey_StatusDown",
            "hotkey_DeleteTerm",
        ],
        "Translate": [
            "hotkey_TranslateSentence",
            "hotkey_TranslatePara",
            "hotkey_TranslatePage",
        ],
        "Copy": [
            "hotkey_CopySentence",
            "hotkey_CopyPara",
            "hotkey_CopyPage",
        ],
        "Misc": [
            "hotkey_Bookmark",
            "hotkey_EditPage",
            "hotkey_NextTheme",
            "hotkey_ToggleHighlight",
            "hotkey_ToggleFocus",
            "hotkey_SaveTerm",
        ],
    }

    settings = {h.key: h.value for h in db.session.query(UserSetting).all()}
    return {
        category: {k: settings[k] for k in keylist}
        for category, keylist in categorized_settings.items()
    }


@bp.route("/shortcuts", methods=["GET", "POST"])
def edit_shortcuts():
    "Edit shortcuts."
    repo = UserSettingRepository(db.session)
    form = UserShortcutsForm()
    if form.validate_on_submit():
        # print(request.form, flush=True)
        # Update the settings in the database
        for k, v in request.form.items():
            # print(f"{k} = {v}", flush=True)
            repo.set_value(k, v)
        db.session.commit()
        refresh_global_settings(db.session)
        flash("Shortcuts updated", "success")
        return redirect("/")

    categorized_settings = _get_categorized_hotkeys()

    setting_descs = {
        "hotkey_StartHover": "Deselect all words",
        "hotkey_PrevWord": "Move to previous word",
        "hotkey_NextWord": "Move to next word",
        "hotkey_PrevUnknownWord": "Move to previous unknown word",
        "hotkey_NextUnknownWord": "Move to next unknown word",
        "hotkey_PrevSentence": "Move to previous sentence",
        "hotkey_NextSentence": "Move to next sentence",
        "hotkey_StatusUp": "Bump the status up by 1",
        "hotkey_StatusDown": "Bump that status down by 1",
        "hotkey_Bookmark": "Bookmark the current page",
        "hotkey_CopySentence": "Copy the sentence of the current word",
        "hotkey_CopyPara": "Copy the paragraph of the current word",
        "hotkey_CopyPage": "Copy the full page",
        "hotkey_TranslateSentence": "Translate the sentence of the current word",
        "hotkey_TranslatePara": "Translate the paragraph of the current word",
        "hotkey_TranslatePage": "Translate the full page",
        "hotkey_NextTheme": "Change to the next theme",
        "hotkey_ToggleHighlight": "Toggle highlights",
        "hotkey_ToggleFocus": "Toggle focus mode",
        "hotkey_Status1": "Set status to 1",
        "hotkey_Status2": "Set status to 2",
        "hotkey_Status3": "Set status to 3",
        "hotkey_Status4": "Set status to 4",
        "hotkey_Status5": "Set status to 5",
        "hotkey_StatusIgnore": "Set status to Ignore",
        "hotkey_StatusWellKnown": "Set status to Well Known",
        "hotkey_DeleteTerm": "Delete term (set status to Unknown)",
        "hotkey_EditPage": "Edit the current page",
        "hotkey_SaveTerm": "Save term in term form",
    }

    return render_template(
        "settings/shortcuts.html",
        setting_descs=setting_descs,
        categorized_settings=categorized_settings,
    )



"""
book helper routines.
"""

import os
from io import StringIO
from datetime import datetime

# pylint: disable=unused-import
from tempfile import TemporaryFile, SpooledTemporaryFile
import requests
from bs4 import BeautifulSoup
from flask import current_app, flash
from openepub import Epub, EpubError
from pypdf import PdfReader
from subtitle_parser import SrtParser, WebVttParser
from werkzeug.utils import secure_filename
from lute.book.model import Book


class BookImportException(Exception):
    """
    Exception to throw on book import error.
    """

    def __init__(self, message="A custom error occurred", cause=None):
        self.cause = cause
        self.message = message
        super().__init__(message)


class Service:
    "Service."

    def _secure_unique_fname(self, filename):
        """
        Return secure name pre-pended with datetime string.
        """
        current_datetime = datetime.now()
        formatted_datetime = current_datetime.strftime("%Y%m%d_%H%M%S")
        f = "_".join([formatted_datetime, secure_filename(filename)])
        return f

    def save_audio_file(self, audio_file_field_data):
        """
        Save the file to disk, return its filename.
        """
        filename = self._secure_unique_fname(audio_file_field_data.filename)
        fp = os.path.join(current_app.env_config.useraudiopath, filename)
        audio_file_field_data.save(fp)
        return filename

    def get_file_content(self, filefielddata):
        """
        Get the content of the file.
        """
        content = None
        _, ext = os.path.splitext(filefielddata.filename)
        ext = (ext or "").lower()
        if ext == ".txt":
            content = self.get_textfile_content(filefielddata)
        if ext == ".epub":
            content = self.get_epub_content(filefielddata)
        if ext == ".pdf":
            msg = """
            Note: pdf imports can be inaccurate, due to how PDFs are encoded.
            Please be aware of this while reading.
            """
            flash(msg, "notice")
            content = self.get_pdf_content_from_form(filefielddata)
        if ext == ".srt":
            content = self.get_srt_content(filefielddata)
        if ext == ".vtt":
            content = self.get_vtt_content(filefielddata)

        if content is None:
            raise ValueError(f'Unknown file extension "{ext}"')
        if content.strip() == "":
            raise BookImportException(f"{filefielddata.filename} is empty.")
        return content

    def get_textfile_content(self, filefielddata):
        "Get content as a single string."
        content = ""
        try:
            content = filefielddata.read()
            return str(content, "utf-8")
        except UnicodeDecodeError as e:
            f = filefielddata.filename
            msg = f"{f} is not utf-8 encoding, please convert it to utf-8 first (error: {str(e)})"
            raise BookImportException(message=msg, cause=e) from e

    def get_epub_content(self, epub_file_field_data):
        """
        Get the content of the epub as a single string.
        """
        content = ""
        try:
            if hasattr(epub_file_field_data.stream, "seekable"):
                epub = Epub(stream=epub_file_field_data.stream)
                content = epub.get_text()
            else:
                # We get a SpooledTemporaryFile from the form but this doesn't
                # implement all file-like methods until python 3.11. So we need
                # to rewrite it into a TemporaryFile
                with TemporaryFile() as tf:
                    epub_file_field_data.stream.seek(0)
                    tf.write(epub_file_field_data.stream.read())
                    epub = Epub(stream=tf)
                    content = epub.get_text()
        except EpubError as e:
            msg = f"Could not parse {epub_file_field_data.filename} (error: {str(e)})"
            raise BookImportException(message=msg, cause=e) from e
        return content

    def get_pdf_content_from_form(self, pdf_file_field_data):
        "Get content as a single string from a PDF file using PyPDF2."
        content = ""
        try:
            pdf_reader = PdfReader(pdf_file_field_data)

            for page in pdf_reader.pages:
                content += page.extract_text()

            return content
        except Exception as e:
            msg = f"Could not parse {pdf_file_field_data.filename} (error: {str(e)})"
            raise BookImportException(message=msg, cause=e) from e

    def get_srt_content(self, srt_file_field_data):
        """
        Get the content of the srt as a single string.
        """
        content = ""
        try:
            srt_content = srt_file_field_data.read().decode("utf-8-sig")

            parser = SrtParser(StringIO(srt_content))
            parser.parse()

            content = "\n".join(subtitle.text for subtitle in parser.subtitles)

            return content
        except Exception as e:
            msg = f"Could not parse {srt_file_field_data.filename} (error: {str(e)})"
            raise BookImportException(message=msg, cause=e) from e

    def get_vtt_content(self, vtt_file_field_data):
        """
        Get the content of the vtt as a single string.
        """
        content = ""
        try:
            vtt_content = vtt_file_field_data.read().decode("utf-8-sig")

            # Check if it is from YouTube
            lines = vtt_content.split("\n")
            if lines[1].startswith("Kind:") and lines[2].startswith("Language:"):
                vtt_content = "\n".join(lines[:1] + lines[3:])

            parser = WebVttParser(StringIO(vtt_content))
            parser.parse()

            content = "\n".join(subtitle.text for subtitle in parser.subtitles)

            return content
        except Exception as e:
            msg = f"Could not parse {vtt_file_field_data.filename} (error: {str(e)})"
            raise BookImportException(message=msg, cause=e) from e

    def book_from_url(self, url):
        "Parse the url and load a new Book."
        s = None
        try:
            timeout = 20  # seconds
            response = requests.get(url, timeout=timeout)
            response.raise_for_status()
            s = response.text
        except requests.exceptions.RequestException as e:
            msg = f"Could not parse {url} (error: {str(e)})"
            raise BookImportException(message=msg, cause=e) from e

        soup = BeautifulSoup(s, "html.parser")
        extracted_text = []

        # Add elements in order found.
        for element in soup.descendants:
            if element.name in ("h1", "h2", "h3", "h4", "p"):
                extracted_text.append(element.text)

        title_node = soup.find("title")
        orig_title = title_node.string if title_node else url

        short_title = orig_title[:150]
        if len(orig_title) > 150:
            short_title += " ..."

        b = Book()
        b.title = short_title
        b.source_uri = url
        b.text = "\n\n".join(extracted_text)
        return b



"""
Text mapping checks.
"""

from datetime import datetime
from lute.models.book import Book, Text, TextBookmark, WordsRead
from lute.db import db
from tests.dbasserts import assert_record_count_equals


def test_save_text_sentences_replaced_in_db(empty_db, english):
    """
    Sentences should only be generated when a Text is saved with the ReadDate saved.
    Sentences are only used for reference lookups.
    """
    b = Book("hola", english)
    t = Text(b, "Tienes un perro. Un gato.")

    db.session.add(t)
    db.session.commit()
    assert_record_count_equals("sentences", 0, "no sentences")

    t.read_date = datetime.now()
    db.session.add(t)
    db.session.commit()
    assert_record_count_equals("sentences", 2, "2 sentences")

    t.text = "Tengo un coche."
    db.session.add(t)
    db.session.commit()
    assert_record_count_equals("sentences", 1, "back to 1 sentences")


def test_delete_text_cascade_deletes_bookmarks_leaves_wordsread(empty_db, english):
    """
    Texts should be able to be deleted even if they have bookmarks.
    In addition, all associated TextBookmark(s) should be deleted when their
    associated Text entry is deleted.
    """
    b = Book("hola", english)
    t = Text(b, "Tienes un perro. Un gato.")
    tb = TextBookmark(text=t, title="Marcador")
    db.session.add(t)
    db.session.add(tb)
    db.session.commit()

    wr = WordsRead(t, datetime.now(), 42)
    db.session.add(wr)
    db.session.commit()

    assert_record_count_equals("texts", 1, "1 text")
    assert_record_count_equals("textbookmarks", 1, "1 bookmark")
    assert_record_count_equals("wordsread", 1, "1 read")

    db.session.delete(t)
    db.session.commit()

    assert_record_count_equals("texts", 0, "0 texts")
    assert_record_count_equals("textbookmarks", 0, "0 bookmarks")
    assert_record_count_equals("wordsread", 1, "still 1 read")
    assert_record_count_equals(
        "select * from wordsread where wrtxid is null", 1, "nulled"
    )



"""
DataTables sqlite tests.
"""

from lute.db import db
from lute.utils.formutils import language_choices, valid_current_language_id
from lute.models.repositories import UserSettingRepository

# pylint: disable=unused-argument


def test_language_choices(app_context, spanish, english):
    "Gets all languages."
    choices = language_choices(db.session)
    assert choices[0][1] == "-", "- at the top"
    langnames = [c[1] for c in choices]
    assert "Spanish" in langnames, "have Spanish"


def test_language_choices_if_only_single_language_exists(app_context, spanish):
    "Gets all languages."
    choices = language_choices(db.session)
    assert choices[0][1] == "Spanish", "sole choice possible"


def test_valid_current_language_id(app_context, spanish, english):
    "Sanity check only."
    repo = UserSettingRepository(db.session)
    repo.set_value("current_language_id", 9999)
    assert int(repo.get_value("current_language_id")) == 9999, "pre-check"
    curr_lang_id = int(valid_current_language_id(db.session))
    assert curr_lang_id == 0, "set back to 0"
    assert int(repo.get_value("current_language_id")) == 0, "re-set to 0"



"""
Book entity.
"""

import sqlite3
from contextlib import closing
from lute.db import db
from lute.parse.base import SentenceGroupIterator

booktags = db.Table(
    "booktags",
    db.Model.metadata,
    db.Column("BtT2ID", db.Integer, db.ForeignKey("tags2.T2ID")),
    db.Column("BtBkID", db.Integer, db.ForeignKey("books.BkID")),
)


class BookTag(db.Model):
    "Term tags."
    __tablename__ = "tags2"

    id = db.Column("T2ID", db.Integer, primary_key=True)
    text = db.Column("T2Text", db.String(20))
    comment = db.Column("T2Comment", db.String(200))

    @staticmethod
    def make_book_tag(text, comment=""):
        "Create a BookTag."
        tt = BookTag()
        tt.text = text
        tt.comment = comment
        return tt


class Book(
    db.Model
):  # pylint: disable=too-few-public-methods, too-many-instance-attributes
    """
    Book entity.
    """

    __tablename__ = "books"

    id = db.Column("BkID", db.SmallInteger, primary_key=True)
    title = db.Column("BkTitle", db.String(length=200))
    language_id = db.Column(
        "BkLgID", db.Integer, db.ForeignKey("languages.LgID"), nullable=False
    )
    source_uri = db.Column("BkSourceURI", db.String(length=1000))
    current_tx_id = db.Column("BkCurrentTxID", db.Integer, default=0)
    archived = db.Column("BkArchived", db.Boolean, default=False)

    audio_filename = db.Column("BkAudioFilename", db.String)
    audio_current_pos = db.Column("BkAudioCurrentPos", db.Float)
    audio_bookmarks = db.Column("BkAudioBookmarks", db.String)

    language = db.relationship("Language")
    texts = db.relationship(
        "Text",
        back_populates="book",
        order_by="Text.order",
        cascade="all, delete-orphan",
    )
    book_tags = db.relationship("BookTag", secondary="booktags")

    def __init__(self, title=None, language=None, source_uri=None):
        self.title = title
        self.language = language
        self.source_uri = source_uri
        self.texts = []
        self.book_tags = []

    def __repr__(self):
        return f"<Book {self.id} {self.title}>"

    def remove_all_book_tags(self):
        self.book_tags = []

    def add_book_tag(self, book_tag):
        if book_tag not in self.book_tags:
            self.book_tags.append(book_tag)

    def remove_book_tag(self, book_tag):
        self.book_tags.remove(book_tag)

    @property
    def page_count(self):
        return len(self.texts)

    def page_in_range(self, n):
        "Return page number that is in the book's page count."
        ret = max(n, 1)
        ret = min(ret, self.page_count)
        return ret

    def text_at_page(self, n):
        "Return the text object at page n."
        pagenum = self.page_in_range(n)
        return self.texts[pagenum - 1]

    def _add_page(self, new_pagenum):
        "Add new page, increment other page orders."
        pages_after = [t for t in self.texts if t.order >= new_pagenum]
        for t in pages_after:
            t.order = t.order + 1
        t = Text(None, "", new_pagenum)
        # TODO fix_refs: None first arg is garbage code.  Passing self
        # as the text's book causes a "SAWarning: Object of type
        # <Text> not in session, add operation along 'Book.texts' will
        # not proceed" warning ... so adding the text to the book
        # manually is needed.  The book's language is required to
        # correctly parse the Text's text though ...
        self.texts.append(t)
        return t

    def add_page_before(self, pagenum):
        "Add page before page n, renumber all subsequent pages, return new page."
        return self._add_page(self.page_in_range(pagenum))

    def add_page_after(self, pagenum):
        "Add page after page n, renumber all subsequent pages, return new page."
        return self._add_page(self.page_in_range(pagenum) + 1)

    def remove_page(self, pagenum):
        "Remove page, renumber all subsequent pages."
        # Don't delete page of single-page books.
        if len(self.texts) == 1:
            return
        texts = [t for t in self.texts if t.order == pagenum]
        if len(texts) == 0:
            return
        texts[0].book = None
        pages_after = [t for t in self.texts if t.order > pagenum]
        for t in pages_after:
            t.order = t.order - 1

    @property
    def is_supported(self):
        "True if the book's language's parser is supported."
        return self.language.is_supported

    @staticmethod
    def create_book(title, language, fulltext, max_word_tokens_per_text=250):
        """
        Create a book with given fulltext content,
        splitting the content into separate Text objects with max
        token count.
        """

        def split_text_at_page_breaks(txt):
            "Break fulltext manually at lines consisting of '---' only."
            # Tried doing this with a regex without success.
            segments = []
            current_segment = ""
            for line in txt.split("\n"):
                if line.strip() == "---":
                    segments.append(current_segment.strip())
                    current_segment = ""
                else:
                    current_segment += line + "\n"
            if current_segment:
                segments.append(current_segment.strip())
            return segments

        pages = []
        for segment in split_text_at_page_breaks(fulltext):
            tokens = language.parser.get_parsed_tokens(segment, language)
            it = SentenceGroupIterator(tokens, max_word_tokens_per_text)
            while toks := it.next():
                s = (
                    "".join([t.token for t in toks])
                    .replace("\r", "")
                    .replace("¶", "\n")
                    .strip()
                )
                pages.append(s)
        pages = [p for p in pages if p.strip() != ""]

        b = Book(title, language)
        for index, page in enumerate(pages):
            t = Text(b, page, index + 1)

        return b


# TODO zzfuture fix: rename class and table to Page/pages
class Text(db.Model):
    """
    Each page in a Book.
    """

    __tablename__ = "texts"

    id = db.Column("TxID", db.Integer, primary_key=True)
    _text = db.Column("TxText", db.String, nullable=False)
    order = db.Column("TxOrder", db.Integer)
    _read_date = db.Column("TxReadDate", db.DateTime, nullable=True)
    bk_id = db.Column("TxBkID", db.Integer, db.ForeignKey("books.BkID"), nullable=False)
    word_count = db.Column("TxWordCount", db.Integer, nullable=True)

    book = db.relationship("Book", back_populates="texts")
    bookmarks = db.relationship(
        "TextBookmark",
        back_populates="text",
        cascade="all, delete-orphan",
    )
    sentences = db.relationship(
        "Sentence",
        back_populates="text",
        order_by="Sentence.order",
        cascade="all, delete-orphan",
    )

    def __init__(self, book, text, order=1):
        self.book = book
        self.text = text
        self.order = order
        self.sentences = []

    @property
    def title(self):
        """
        Text title is the book title + page fraction.
        """
        b = self.book
        s = f"({self.order}/{self.book.page_count})"
        t = f"{b.title} {s}"
        return t

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, s):
        self._text = s
        if s.strip() == "":
            return
        toks = self._get_parsed_tokens()
        wordtoks = [t for t in toks if t.is_word]
        self.word_count = len(wordtoks)
        if self._read_date is not None:
            self._load_sentences_from_tokens(toks)

    @property
    def read_date(self):
        return self._read_date

    @read_date.setter
    def read_date(self, s):
        self._read_date = s
        # Ensure loaded.
        self.load_sentences()

    def _get_parsed_tokens(self):
        "Return the tokens."
        lang = self.book.language
        return lang.parser.get_parsed_tokens(self.text, lang)

    def _load_sentences_from_tokens(self, parsedtokens):
        "Save sentences using the tokens."
        parser = self.book.language.parser
        self._remove_sentences()
        curr_sentence_tokens = []
        sentence_num = 1

        def _add_current():
            "Create and add sentence from current state."
            if curr_sentence_tokens:
                se = Sentence.from_tokens(curr_sentence_tokens, parser, sentence_num)
                self._add_sentence(se)
            # Reset for the next sentence.
            curr_sentence_tokens.clear()

        for pt in parsedtokens:
            curr_sentence_tokens.append(pt)
            if pt.is_end_of_sentence:
                _add_current()
                sentence_num += 1

        # Add any stragglers.
        _add_current()

    def load_sentences(self):
        """
        Parse the current text and create Sentence objects.
        """
        toks = self._get_parsed_tokens()
        self._load_sentences_from_tokens(toks)

    def _add_sentence(self, sentence):
        "Add a sentence to the Text."
        if sentence not in self.sentences:
            self.sentences.append(sentence)
            sentence.text = self

    def _remove_sentences(self):
        "Remove all sentence from the Text."
        for sentence in self.sentences:
            sentence.text = None
        self.sentences = []


class WordsRead(db.Model):
    """
    Tracks reading events for Text entities.
    """

    __tablename__ = "wordsread"
    id = db.Column("WrID", db.Integer, primary_key=True)
    language_id = db.Column(
        "WrLgID", db.Integer, db.ForeignKey("languages.LgID"), nullable=False
    )
    tx_id = db.Column(
        "WrTxID",
        db.Integer,
        db.ForeignKey("texts.TxID", ondelete="SET NULL"),
        nullable=True,
    )
    read_date = db.Column("WrReadDate", db.DateTime, nullable=False)
    word_count = db.Column("WrWordCount", db.Integer, nullable=False)

    def __init__(self, text, read_date, word_count):
        self.tx_id = text.id
        self.language_id = text.book.language.id
        self.read_date = read_date
        self.word_count = word_count


class Sentence(db.Model):
    """
    Parsed sentences for a given Text.

    The Sentence contains the parsed tokens, joined by the zero-width string.
    """

    __tablename__ = "sentences"

    id = db.Column("SeID", db.Integer, primary_key=True)
    tx_id = db.Column("SeTxID", db.Integer, db.ForeignKey("texts.TxID"), nullable=False)
    order = db.Column("SeOrder", db.Integer, default=1)
    text_content = db.Column("SeText", db.Text, default="")
    textlc_content = db.Column("SeTextLC", db.Text)

    text = db.relationship("Text", back_populates="sentences")

    def set_lowercase_text(self, parser):
        """
        Load textlc_content from text_content.

        If a call to sqlite's LOWER() function for the text_content
        returns the same lowercase text as a call to the parser,
        store '*' as the lowercase text.  This seeming hack can save a
        pile of space: for my ~30meg db of ~135K sentences, only 750
        sentences were different when lowercased by the LOWER() vs by
        the parser.

        This method is public for use in the data_cleanup module.
        """

        def _get_sql_lower(input_string):
            "Returns result of sqlite LOWER call of input_string."
            if input_string is None:
                return None
            with sqlite3.connect(":memory:") as conn, closing(conn.cursor()) as cur:
                cur.execute("SELECT LOWER(?)", (input_string,))
                result = cur.fetchone()
                return result[0]

        lcased = parser.get_lowercase(self.text_content)
        if lcased == _get_sql_lower(self.text_content):
            lcased = "*"
        self.textlc_content = lcased

    @staticmethod
    def from_tokens(tokens, parser, senumber):
        """
        Create a new Sentence from ParsedTokens.
        """

        def _sentence_string(string_array):
            "Create properly-zws-joined sentence string."
            zws = chr(0x200B)  # Zero-width space.
            s = zws.join(string_array).strip(" ")
            # The zws is added at the start and end of each
            # sentence, to standardize the string search when
            # looking for terms.
            return zws + s + zws

        sentence = Sentence()
        sentence.order = senumber
        sentence.text_content = _sentence_string([t.token for t in tokens])
        sentence.set_lowercase_text(parser)
        return sentence


class TextBookmark(db.Model):
    """
    Bookmarks for a given Book page

    The TextBookmark includes a title
    """

    __tablename__ = "textbookmarks"

    id = db.Column("TbID", db.Integer, primary_key=True)
    tx_id = db.Column(
        "TbTxID",
        db.Integer,
        db.ForeignKey("texts.TxID", ondelete="CASCADE"),
        nullable=False,
    )
    title = db.Column("TbTitle", db.Text, nullable=False)

    text = db.relationship("Text", back_populates="bookmarks")


class BookStats(db.Model):
    "The stats table."
    __tablename__ = "bookstats"

    BkID = db.Column(db.Integer, primary_key=True)
    distinctterms = db.Column(db.Integer)
    distinctunknowns = db.Column(db.Integer)
    unknownpercent = db.Column(db.Integer)
    status_distribution = db.Column(db.String, nullable=True)



"""
Show bookmarks in datatables.
"""

from lute.utils.data_tables import DataTablesSqliteQuery


def get_data_tables_list(parameters, book_id, session):
    "Bookmark json data for datatables."

    base_sql = f"""
      SELECT tb.TbID, tb.TbTxID, tb.TbTitle, tx.TxOrder
      FROM textbookmarks as tb
      INNER JOIN texts as tx ON tb.TbTxID = tx.TxID
      WHERE tx.TxBkID = { book_id }
    """

    connection = session.connection()
    return DataTablesSqliteQuery.get_data(base_sql, parameters, connection)



"""
DataTables sqlite tests.
"""

import re
import pytest
from lute.utils.data_tables import DataTablesSqliteQuery

# pylint: disable=line-too-long
# The tests all generate sql, which in most cases is quite long.


@pytest.fixture(name="basesql")
def fixture_basesql():
    "Base select for tests."
    return "select CatID, Color, Food from Cats"


@pytest.fixture(name="columns")
def fixture_columns():
    "Columns that would be returned from DataTablesFlaskParamParser."
    return [
        {"data": "0", "name": "CatID", "searchable": False, "orderable": False},
        {"data": "1", "name": "Color", "searchable": True, "orderable": True},
        {"data": "2", "name": "Food", "searchable": True, "orderable": True},
    ]


@pytest.fixture(name="parameters")
def fixture_parameters(columns):
    "Parameters that would be passed from DataTablesFlaskParamParser."
    return {
        "draw": "1",
        "columns": columns,
        "order": [{"column": "1", "dir": "asc"}],
        "start": "10",
        "length": "50",
        "search": {"value": "", "regex": False},
    }


def test_smoke_test(basesql, parameters):
    "Thing works when params are passed!"
    actual = DataTablesSqliteQuery.get_sql(basesql, parameters)
    expected = {
        "recordsTotal": "select count(*) from (select CatID, Color, Food from Cats) realbase",
        "recordsFiltered": "select count(*) from (select CatID, Color, Food from Cats) realbase ",
        "data": "SELECT * FROM (select * from (select CatID, Color, Food from Cats) realbase  ORDER BY Color asc, Color, Food LIMIT 10, 50) src ORDER BY Color asc, Color, Food",
        "params": {},
        "draw": 1,
    }
    assert actual == expected


def test_sorting(basesql, parameters):
    "Sorting is included in the data query."
    parameters["order"][0]["column"] = "2"
    parameters["order"][0]["dir"] = "desc"

    actual = DataTablesSqliteQuery.get_sql(basesql, parameters)
    expected = {
        "recordsTotal": "select count(*) from (select CatID, Color, Food from Cats) realbase",
        "recordsFiltered": "select count(*) from (select CatID, Color, Food from Cats) realbase ",
        "data": "SELECT * FROM (select * from (select CatID, Color, Food from Cats) realbase  ORDER BY Food desc, Color, Food LIMIT 10, 50) src ORDER BY Food desc, Color, Food",
        "params": {},
        "draw": 1,
    }
    assert actual == expected


def test_single_search(basesql, parameters):
    "A single search param is used in the where clause."
    parameters["search"]["value"] = "XXX"

    actual = DataTablesSqliteQuery.get_sql(basesql, parameters)

    expected = {
        "recordsTotal": "select count(*) from (select CatID, Color, Food from Cats) realbase",
        "recordsFiltered": "select count(*) from (select CatID, Color, Food from Cats) realbase WHERE (Color LIKE '%' || :s0 || '%' OR Food LIKE '%' || :s0 || '%')",
        "data": "SELECT * FROM (select * from (select CatID, Color, Food from Cats) realbase WHERE (Color LIKE '%' || :s0 || '%' OR Food LIKE '%' || :s0 || '%') ORDER BY Color asc, Color, Food LIMIT 10, 50) src ORDER BY Color asc, Color, Food",
        "params": {"s0": "XXX"},
        "draw": 1,
    }
    assert actual == expected


def test_multiple_search_terms(basesql, parameters):
    "Different search terms are used for all the searchable columns."
    parameters["search"]["value"] = "XXX YYY"

    actual = DataTablesSqliteQuery.get_sql(basesql, parameters)

    expected = {
        "recordsTotal": "select count(*) from (select CatID, Color, Food from Cats) realbase",
        "recordsFiltered": "select count(*) from (select CatID, Color, Food from Cats) realbase WHERE (Color LIKE '%' || :s0 || '%' OR Food LIKE '%' || :s0 || '%') AND (Color LIKE '%' || :s1 || '%' OR Food LIKE '%' || :s1 || '%')",
        "data": "SELECT * FROM (select * from (select CatID, Color, Food from Cats) realbase WHERE (Color LIKE '%' || :s0 || '%' OR Food LIKE '%' || :s0 || '%') AND (Color LIKE '%' || :s1 || '%' OR Food LIKE '%' || :s1 || '%') ORDER BY Color asc, Color, Food LIMIT 10, 50) src ORDER BY Color asc, Color, Food",
        "params": {"s0": "XXX", "s1": "YYY"},
        "draw": 1,
    }
    assert actual == expected


def test_search_regex_markers(basesql, parameters):
    "^ or $ has special meaning in the searches."

    def assert_where_equals(search_string, expected):
        parameters["search"]["value"] = search_string
        actual = DataTablesSqliteQuery.get_sql(basesql, parameters)
        filtered = actual["recordsFiltered"]
        where = re.sub(".* WHERE ", "", filtered)
        assert where == expected, search_string

    assert_where_equals(
        "XXX", "(Color LIKE '%' || :s0 || '%' OR Food LIKE '%' || :s0 || '%')"
    )
    assert_where_equals(
        "^XXX", "(Color LIKE '' || :s0 || '%' OR Food LIKE '' || :s0 || '%')"
    )
    assert_where_equals(
        "XXX$", "(Color LIKE '%' || :s0 || '' OR Food LIKE '%' || :s0 || '')"
    )
    assert_where_equals(
        "^XXX$", "(Color LIKE '' || :s0 || '' OR Food LIKE '' || :s0 || '')"
    )
